﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using OpenTK.Graphics.OpenGL;
using OpenTK.Graphics;
using System.IO;

namespace Triangle
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        Color4 backgroundColor = Color4.Black;

        string infoFileName = "info.txt";

        string vShaderSource =
            "#version 330\n" +
            "in vec4 a_Position;\n" +
            "void main() {\n" +
            "gl_Position = a_Position;\n" +
            "}\n";
        string fShaderSource =
            "#version 330\n" +
            "out vec4 outColor;\n" +
            "void main() {\n" +
            "outColor = vec4(1.0, 0.0, 0.0, 1.0);\n" +
            "}\n";

        int program;

        int n;
        bool canDraw = false;

        public MainWindow()
        {
            InitializeComponent();
        }

        private void WindowsFormsHost_Initialized(object sender, EventArgs e)
        {
            renderCanvas.MakeCurrent();
        }

        private void renderCanvas_Load(object sender, EventArgs e)
        {
            if (File.Exists(infoFileName))
            {
                // Clear File
                File.WriteAllText(infoFileName, "");
            }

            // Load shaders from files
            //LoadShader("./Shaders/vShader.glsl", out vShaderSource);
            //LoadShader("./Shaders/fShader.glsl", out fShaderSource);
            //if (vShaderSource == null || fShaderSource == null)
            //{
            //    return;
            //}

            // Initilize the shaders
            if (!InitShaders(vShaderSource, fShaderSource, out program))
            {
                File.AppendAllText(infoFileName, "Failed to intialize shaders");
                return;
            }

            // Set the vertex information
            n = InitVertexBuffers();
            if (n < 0)
            {
                File.AppendAllText(infoFileName, "Failed to set the vertex information");
                return;
            }

            // Specify the color for clearing
            GL.ClearColor(backgroundColor);
            canDraw = true;
        }

        private int InitVertexBuffers()
        {
            float[] verticesColors = new float[]
            {
                // Vertex coordinates and color
                0.0f,  0.5f,
                -0.5f, -0.5f,
                0.5f, -0.5f,
            };
            int n = 3;

            // Create a buffer object
            int vertexColorBuffer;
            GL.GenBuffers(1, out vertexColorBuffer);

            // Bind the buffer object to target
            GL.BindBuffer(BufferTarget.ArrayBuffer, vertexColorBuffer);
            GL.BufferData(BufferTarget.ArrayBuffer, (IntPtr)(verticesColors.Length * sizeof(float)), verticesColors, BufferUsageHint.StaticDraw);

            // Get the storage location of a_Position, assign and enable buffer
            int a_Position = GL.GetAttribLocation(program, "a_Position");
            if (a_Position < 0)
            {
                File.AppendAllText(infoFileName, "Failed to get the storage location of a_Position");
                return -1;
            }
            GL.VertexAttribPointer(a_Position, 2, VertexAttribPointerType.Float, false, 0, 0);
            GL.EnableVertexAttribArray(a_Position);

            //// Get the storage location of a_Position, assign buffer and enable
            //int a_Color = GL.GetAttribLocation(program, "a_Color");
            //if (a_Color < 0)
            //{
            //    File.AppendAllText(infoFileName, "Failed to get the storage location of a_Color");
            //    return -1;
            //}
            //GL.VertexAttribPointer(a_Color, 3, VertexAttribPointerType.Float, false, sizeof(float) * 5, sizeof(float) * 2);
            //GL.EnableVertexAttribArray(a_Color);

            //// Unbind the buffer object
            //GL.BindBuffer(BufferTarget.ArrayBuffer, 0);

            return n;
        }

        private void renderCanvas_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
        {
            GL.Viewport(0, 0, renderCanvas.Width, renderCanvas.Height);
            GL.Clear(ClearBufferMask.ColorBufferBit);

            Draw();

            GL.Flush();
            renderCanvas.SwapBuffers();
        }

        private void Draw()
        {
            if (!canDraw) return;
            DrawTriangle();
        }

        private void DrawTriangle()
        {
            GL.DrawArrays(PrimitiveType.Triangles, 0, n);
        }

        private void SetBackgroundColor(Color4 color)
        {
            backgroundColor = color;
            GL.ClearColor(backgroundColor);
            renderCanvas.Invalidate();
        }

        private void SetBackgroundColorButton_Click(object sender, RoutedEventArgs e)
        {
            using (var dialog = new System.Windows.Forms.ColorDialog())
            {
                if (dialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                {
                    SetBackgroundColor(dialog.Color);
                }
            }
        }

        private void SetTriangleColorButton_Click(object sender, RoutedEventArgs e)
        {

        }

        ///<summary>
        ///Create a program object and make current
        ///</summary>
        ///<param name="vShader">a vertex shader program</param>
        ///<param name="fShader">a fragment shader program</param>
        ///<param name="program">created program</param>
        ///<returns>
        ///return true, if the program object was created and successfully made current
        ///</returns>
        private bool InitShaders(string vShader, string fShader, out int program)
        {
            program = CreateProgram(vShader, fShader);
            if (program == 0)
            {
                File.AppendAllText(infoFileName, "Failed to create program");
                return false;
            }

            GL.UseProgram(program);

            return true;
        }

        private int CreateProgram(string vShader, string fShader)
        {
            // Create shader object
            int vertexShader = LoadShader(ShaderType.VertexShader, vShader);
            int fragmentShader = LoadShader(ShaderType.FragmentShader, fShader);
            if (vertexShader == 0 || fragmentShader == 0)
            {
                return 0;
            }

            // Create a program object
            int program = GL.CreateProgram();
            if (program == 0)
            {
                return 0;
            }

            // Attach the shader objects
            GL.AttachShader(program, vertexShader);
            GL.AttachShader(program, fragmentShader);

            // Link the program object
            GL.LinkProgram(program);

            // Check the result of linking
            int status;
            GL.GetProgram(program, GetProgramParameterName.LinkStatus, out status);
            if (status == 0)
            {
                string errorString = string.Format("Failed to link program: {0}", GL.GetProgramInfoLog(program));
                File.AppendAllText(infoFileName, errorString);
                GL.DeleteProgram(program);
                GL.DeleteShader(vertexShader);
                GL.DeleteShader(fragmentShader);
                return 0;
            }

            return program;
        }

        private int LoadShader(ShaderType shaderType, string shaderSource)
        {
            // Create shader object
            int shader = GL.CreateShader(shaderType);
            if (shader == 0)
            {
                File.AppendAllText(infoFileName, "Unable to create shader");
                return 0;
            }

            // Set the shader program
            GL.ShaderSource(shader, shaderSource);

            // Compile the shader
            GL.CompileShader(shader);

            // Check the result of compilation
            int status;
            GL.GetShader(shader, ShaderParameter.CompileStatus, out status);
            if (status == 0)
            {
                string errorString = string.Format("Failed to compile {0} shader: {1}", shaderType.ToString(), GL.GetShaderInfoLog(shader));
                File.AppendAllText(infoFileName, errorString);
                GL.DeleteShader(shader);
                return 0;
            }

            return shader;
        }

        private void LoadShader(string fileName, out string shaderSource)
        {
            shaderSource = null;

            using (StreamReader sr = new StreamReader(fileName))
            {
                shaderSource = sr.ReadToEnd();
            }
        }
    }
}
